%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Copyright (c) 2001 by Erik Johansson.  All Rights Reserved 
%% Time-stamp: <01/07/25 14:30:06 happi>
%% ====================================================================
%%  Filename : 	hipe_sparc_ra.erl
%%  Module   :	hipe_sparc_ra
%%  Purpose  :  
%%  Notes    : 
%%  History  :	* 2001-07-20 Erik Johansson (happi@csd.uu.se): 
%%               Created.
%%  CVS      :
%%              $Author: happi $
%%              $Date: 2001/07/25 12:32:33 $
%%              $Revision: 1.3 $
%% ====================================================================
%%  Exports  :
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

-module(hipe_sparc_ra).
-export([allocate/3]).

-include("../main/hipe.hrl").

allocate(Fun, SparcCfg, Options) ->
  ?when_option(time, Options, ?start_timer("Regalloc")),
  ?when_option(count_instrs, Options,
	       put(pre_ra_instrs,
		   get(pre_ra_instrs)+
		   hipe_sparc_size:count_instrs_cfg(SparcCfg))),
  ?when_option(timeregalloc, Options, 
	       put(regalloctime1,erlang:statistics(runtime))),

  NewCfg = 
    case property_lists:get_value(regalloc,Options) of
      linear_scan ->
	hipe_sparc_ra_ls:alloc(SparcCfg, Options);
      graph_color ->
	hipe_sparc_ra_graph_color:alloc(SparcCfg);
      optimistic_ls ->
	hipe_sparc_ra_ols:alloc(SparcCfg, Options);
      coalescing ->
	hipe_sparc_ra_coalescing:alloc( hipe_sparc_multimove:remove_multimoves(SparcCfg), Options);
%	  hipe_sparc_multimove:remove_multimoves(SparcCfg));
      memory ->
	hipe_sparc_ra_memory:alloc(SparcCfg);
      _ -> %% linear_scan made default register allocator
	hipe_sparc_ra_ls:alloc(SparcCfg, Options)
    end,
  ?when_option(time, Options, ?stop_timer("Regalloc done")),
  ?when_option(timeregalloc, Options, put(regalloctime2,erlang:statistics(runtime))),
  ?when_option(timeregalloc, Options, 
	       case
		 {get(regalloctime1),
		  get(regalloctime2)} of 
		 {{T1,_},{T2,_}} ->
		   put(regalloctime,
		       get(regalloctime) + 
		       (T2-T1))
	       end),
  ?when_option(count_instrs, Options,
	       put(post_ra_instrs,
		   get(post_ra_instrs)+
		   hipe_sparc_size:count_instrs_cfg(NewCfg))),
  NewCfg.
