%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Copyright (c) 2001 by Erik Johansson.  All Rights Reserved 
%% -*- erlang-indent-level: 2 -*-
%% ====================================================================
%%  Filename : 	hipe_x86_ra_graph_color.erl
%%  Module   :	hipe_x86_ra_graph_color
%%  Purpose  :  
%%  Notes    : 
%%  History  :	* 2001-07-26 Erik Johansson (happi@csd.uu.se): 
%%               Created.
%%  CVS      :
%%              $Author: happi $
%%              $Date: 2001/07/30 13:58:25 $
%%              $Revision: 1.3 $
%% ====================================================================
%%  Exports  :
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

-module(hipe_x86_ra_graph_color).
-export([ra/2]).
-include("hipe_x86.hrl").
-include("../main/hipe.hrl").

-define(countspills,true).

-ifdef(countspills).
-define(count_spills(X), X).
-define(update_spillsum,
	case get(totalspill) of 
	  {__Loads,__Stores} ->
	    put(totalspill,{__Loads+get(loads),
			    __Stores+get(stores)}),
	    case get(spilledtemps) of
	      __Temps -> 
		put(spilledtemps, __Temps+get(temps));
	      _ ->
		true
	    end;
	  _ -> true 
	end).
-else.
-define(count_spills(X), true).
-define(update_spillsum,true).
-endif.

ra(X86Defun,Options) ->
  ?count_spills({put(loads,0),put(stores,0),put(temps,0)}),
  SpillLimit = hipe_x86_specific:number_of_temporaries(
		 hipe_x86_cfg:init(X86Defun)),
  alloc(X86Defun, 0, SpillLimit, Options).

alloc(X86Defun, SpillIndex, SpillLimit, Options) ->
  X86Cfg = hipe_x86_cfg:init(X86Defun),
  {Coloring, NewSpillIndex} = 
    hipe_graph_coloring_regalloc:regalloc(X86Cfg, SpillIndex, 
				      SpillLimit, hipe_x86_specific ),

  ?count_spills(put(temps, NewSpillIndex)),
  {NewX86Defun, _, DontSpill} =
    hipe_x86_ra_postconditions:check_and_rewrite(X86Defun,
						 Coloring,
						 [],
						 Options),

  case DontSpill of
    [] -> %% No new temps, we are done.
      {NewX86Defun, Coloring};
    _ -> 
      %% Since SpillLimit is used as a low-water-mark
      %% the list of temps not to spill is uninteresting.
      alloc( NewX86Defun, NewSpillIndex, SpillLimit, Options)
  end.



